"""
Tools to run over code that don't lint and don't format

Currently included:
- Pytest
- Upload to pypi (twine)
"""

# Stdlib
import io
import json
import os
import subprocess  # noqa: B404
import sys

# External Libraries
from snekchek.misc import __version__
from snekchek.structure import Linter
from snekchek.utils import redirect_stderr, redirect_stdout


def get_tools():
    return Pytest, Pypi


class Pytest(Linter):
    requires_install = ["pytest", "pytest-json"]

    def run(self, _: list) -> None:
        import pytest

        with redirect_stdout(sys.stderr):
            exitcode = pytest.main(
                ["--json=.log.json", "-qqqq", "-c", self.confpath])
        self.status_code = exitcode

        with open(".log.json") as file:
            data = json.load(file)

        os.remove(".log.json")

        self.hook([
            test for test in data['report']['tests']
            if test['outcome'] == 'failed'
        ])


class Pypi(Linter):
    requires_install = ["twine", "wheel", "requests"]

    def run(self, _: list) -> None:
        import requests
        import twine.commands.upload

        try:
            with redirect_stdout(io.StringIO()), \
                    redirect_stderr(io.StringIO()):
                proc = subprocess.Popen(  # noqa: B603
                    [sys.executable, "setup.py", "sdist", "bdist_wheel"],
                    stdout=subprocess.DEVNULL)
                proc.wait()
                twine.commands.upload.upload(
                    ["dist/*{0}*".format(__version__)],
                    self.conf["TWINE_REPOSITORY"], self.conf.as_bool("sign"),
                    self.conf.get("identity"), self.conf["TWINE_USERNAME"],
                    self.conf["TWINE_PASSWORD"], self.conf.get("comment"),
                    self.conf.get("sign-with"), self.confpath,
                    self.conf.get("skip-existing", True), None, None, None)

        except requests.exceptions.HTTPError as err:
            print(err)

        self.hook([])



"""
Checkers for security issues in the project.

Checkers included:
- Safety
- Dodgy
"""

# Stdlib
import io
import json
import os

# External Libraries
from snekchek.structure import Linter
from snekchek.utils import redirect_stdout


def get_security() -> list:
    return Safety, Dodgy


class Safety(Linter):
    requires_install = ["safety"]

    def run(self, _: list) -> None:
        import safety.cli

        if "requirements.txt" not in os.listdir():
            self.hook([])
            return

        outfile = io.StringIO()

        try:
            with redirect_stdout(outfile):
                safety.cli.check.callback(
                    self.conf.get("pyup_key", ""), self.conf.get(
                        "db_path", ""), True, False, False, False, False,
                    "requirements.txt", self.conf.as_list("ignore"))
        except SystemExit:
            # Raised by safety
            pass

        outfile.seek(0)

        json_data = json.load(outfile)

        self.status_code = 1 if json_data else 0

        self.hook(json_data)


class Dodgy(Linter):
    requires_install = ["dodgy"]

    def run(self, _: list) -> None:
        import dodgy.run

        data = dodgy.run.run_checks(".", self.conf.as_list("ignore_paths"))

        self.status_code = 1 if data else 0

        self.hook(data)



"""
This module provides redirect_* functions for Python <3.5
"""

# Stdlib
import sys


class _RedirectStream:
    _stream = None

    def __init__(self, new_target):
        self._new_target = new_target
        # We use a list of old targets to make this CM re-entrant
        self._old_targets = []

    def __enter__(self):
        self._old_targets.append(getattr(sys, self._stream))
        setattr(sys, self._stream, self._new_target)
        return self._new_target

    def __exit__(self, exctype, excinst, exctb):
        setattr(sys, self._stream, self._old_targets.pop())


class redirect_stdout(_RedirectStream):  # pylint: disable=invalid-name
    """Context manager for temporarily redirecting stdout to another file.
        # How to send help() to stderr
        with redirect_stdout(sys.stderr):
            help(dir)
        # How to write help() to a file
        with open('help.txt', 'w') as f:
            with redirect_stdout(f):
                help(pow)
    """

    _stream = "stdout"


class redirect_stderr(_RedirectStream):  # pylint: disable=invalid-name
    """Context manager for temporarily redirecting stderr to another file."""

    _stream = "stderr"



""" Common classes and utility functions. """

# Stdlib
import json
import os
import re
import subprocess  # noqa: B404
import sys

# External Libraries
import snekchek.format


def flatten(nested_list: list) -> list:
    """Flattens a list, ignore all the lambdas."""
    return list(sorted(filter(lambda y: y is not None,
                              list(map(lambda x: (nested_list.extend(x)  # noqa: T484
                                                  if isinstance(x, list) else x),
                                       nested_list)))))


def get_py_files(dir_name: str) -> list:
    """Get all .py files."""
    return flatten([
        x for x in
        [["{0}/{1}".format(path, f) for f in files if f.endswith(".py")]
         for path, _, files in os.walk(dir_name)
         if not path.startswith("./build")] if x
    ])


class ModuleNotInstalled(Exception):
    pass


class CheckHandler:
    def __init__(self,
                 file: str,
                 out_json: bool,
                 check_dir: str = ".",
                 files=None):
        # Do this here so setup.py doesn't error
        from snekchek.baseconfig import config
        import configobj

        self.parser = config
        self.parser.merge(configobj.ConfigObj(file))

        self.fn = file  # pylint: disable=invalid-name
        self.status_code = 0
        self.logs = {}
        self.current = ''
        self.json = out_json

        self.indent = 4 if "--debug" in sys.argv else None

        self.files = files or get_py_files(check_dir)

        patt = re.compile(r"^(?P<package>\S+?) \((?P<version>\S+)\)$", re.M)

        args = [sys.executable, '-m', 'pip', 'list']

        if sys.version_info >= (3, 4, 0):
            args.extend(['--format=legacy'])

        proc = subprocess.Popen(  # noqa: B603
            args, stdout=subprocess.PIPE)

        proc.wait()

        matches = list(patt.finditer(proc.stdout.read().decode()))

        self.installed = [p.group("package") for p in matches]

    def exit(self) -> None:
        """Raise SystemExit with correct status code and output logs."""
        total = sum(len(logs) for logs in self.logs.values())
        if self.json:
            self.logs['total'] = total
            print(json.dumps(self.logs, indent=self.indent))

        else:
            for name, log in self.logs.items():
                if not log or self.parser[name].as_bool("quiet"):
                    continue

                print("[[{0}]]".format(name))
                getattr(snekchek.format, name + "_format")(log)
                print("\n")

            print("-" * 30)
            print("Total:", total)

        sys.exit(self.status_code)

    def run_linter(self, linter) -> None:
        """Run a checker class"""
        self.current = linter.name

        if (linter.name not in self.parser["all"].as_list("linters")
                or linter.base_pyversion > sys.version_info):  # noqa: W503
            return

        if any(x not in self.installed for x in linter.requires_install):
            raise ModuleNotInstalled(linter.requires_install)

        linter.add_output_hook(self.out_func)
        linter.set_config(self.fn, self.parser[linter.name])
        linter.run(self.files)
        self.status_code = self.status_code or linter.status_code

    def out_func(self, data) -> None:
        self.logs[self.current] = data


class Linter:
    """Common shared class for all linters/stylers/tools"""

    requires_install = []
    base_pyversion = (3, 0, 0)

    def __init__(self):
        self.status_code = 0
        self.hook = None
        self.confpath = None
        self.conf = None

    def add_output_hook(self, func) -> None:
        self.hook = func

    def set_config(self, confpath: str, section) -> None:
        self.confpath = confpath
        self.conf = section

    @property
    def name(self):
        return self.__class__.__name__.lower()



"""
This file contains Style checkers.

Stylers included:
- isort
- yapf
"""

# Stdlib
import io

# External Libraries
from snekchek.structure import Linter
from snekchek.utils import redirect_stdout


def get_stylers() -> list:
    return ISort, Yapf


class ISort(Linter):
    requires_install = ["isort"]

    def run(self, files: list) -> None:
        import isort

        self.conf['line_length'] = self.conf.as_int('line_length')
        self.conf['sections'] = self.conf.as_list('sections')
        self.conf['multi_line_output'] = self.conf.as_int('multi_line_output')

        res = []

        for filename in files:
            with redirect_stdout(io.StringIO()):  # mute stdout
                sort = isort.SortImports(filename, **self.conf)

            if sort.skipped:
                continue

            self.status_code = self.status_code or (1
                                                    if sort.incorrectly_sorted
                                                    else 0)

            if self.conf.as_bool('inplace'):
                with open(filename, "w") as file:
                    file.write(sort.output)

            else:
                with open(filename) as file:
                    out = io.StringIO()
                    with redirect_stdout(out):
                        sort._show_diff(file.read())  # pylint: disable=protected-access
                    out.seek(0)
                    diff = out.read()

                if diff.strip():
                    res.append(diff.strip())

        self.hook(res)


class Yapf(Linter):
    requires_install = ["yapf"]
    base_pyversion = (3, 4, 0)

    def run(self, files: list) -> None:
        import yapf.yapflib.yapf_api

        res = []

        for file in files:
            code, _, changed = yapf.yapflib.yapf_api.FormatFile(
                file, style_config=self.confpath)

            self.status_code = self.status_code or (1 if changed else 0)

            if changed:

                if self.conf.as_bool('inplace'):
                    with open(file, "w") as new_file:
                        new_file.write(code)

                else:
                    res.append(code.strip())

        self.hook(res)



"""Formatting functions for each linter"""


def flake8_format(data: list) -> None:
    for row in data:
        print("{path}:{line}:{col}: {errcode} {msg}".format(**row))


def vulture_format(data: list) -> None:
    for row in data:
        print("{path}:{line}: {err} ({conf}% confidence)".format(**row))


def pylint_format(data: list) -> None:
    last_path = ""
    for row in data:
        if row['path'] != last_path:
            print("File: {0}".format(row['path']))
            last_path = row['path']

        print("{type_}:{line:>3}, {column:>2}: {message} ({symbol})".format(
            type_=row['type'][0].upper(), **row))


def pyroma_format(data: dict) -> None:
    for row in list(data['modules'].values())[0]:
        print(row)


def isort_format(data: list) -> None:
    for diff in data:
        print(diff)


def yapf_format(data: list) -> None:
    for row in data:
        print(row)


def pypi_format(data: list) -> None:
    for row in data:
        print(row)


def safety_format(data: list) -> None:
    for row in data:
        print("[{row[4]}] ({row[0]}{row[1]}) {row[3]}".format(row=row))


def dodgy_format(data: list) -> None:
    for row in data:
        print("{row[1]}:{row[0]}: {row[2]}".format(row=row))


def pytest_format(data: list) -> None:
    for test in data:
        print(test['name'])
        print(test['call']['longrepr'])



"""
This file contains linters.

Linters included:

- flake8
- flake8-bugbear (ext)
- flake8-import-order (ext)
- flake8-mypy (ext)
- flake8-docstrings (ext)
- flake8-todo (ext)
- flake8-requirements (ext)
- flake8-string-format (ext)
- flake8-tidy-import (ext)
- flake8-bandit (ext, bandit)
- pylint
- vulture
- pyroma
"""

# Stdlib
import io
import json
import re

# External Libraries
from snekchek.structure import Linter
from snekchek.utils import redirect_stderr, redirect_stdout


def get_linters() -> list:
    return Vulture, Pylint, Pyroma, Flake8


class Flake8(Linter):
    requires_install = ["flake8"]

    patt = re.compile(r"(?P<path>[^:]+):(?P<line>[0-9]+):(?P<col>[0-9]+): "
                      r"(?P<errcode>[A-Z][0-9]+) (?P<msg>.+)$\n", re.M)

    def run(self, files: list) -> None:
        import flake8.main.cli

        file = io.StringIO()
        with redirect_stdout(file):
            try:
                sett = ["--config=" + self.confpath]
                sett.extend(files)
                flake8.main.cli.main(sett)
            except SystemExit:
                print("aaa")
        file.seek(0)
        matches = list(self.patt.finditer(file.read()))
        self.status_code = 1 if matches else 0
        self.hook(
            list(
                sorted(
                    [x.groupdict() for x in matches],
                    key=lambda x: x["line"])))


class Vulture(Linter):
    requires_install = ["vulture"]

    patt = re.compile(
        r"^(?P<path>[^:]+):(?P<line>[0-9]+): "
        r"(?P<err>unused (class|attribute|function) '[a-zA-Z0-9]+') "
        r"\((?P<conf>[0-9]+)% confidence\)$")

    def run(self, files: list) -> None:
        import vulture.core

        vult = vulture.core.Vulture(self.conf.as_bool('verbose'))
        vult.scavenge(files, [x.strip() for x in self.conf.as_list("exclude")])
        file = io.StringIO()
        with redirect_stdout(file):
            vult.report(
                self.conf.as_int("min-confidence"),
                self.conf.as_bool("sort-by-size"))
        file.seek(0)
        matches = list(self.patt.finditer(file.read()))
        self.status_code = 1 if matches else 0
        self.hook(
            list(
                sorted(
                    [x.groupdict() for x in matches],
                    key=lambda x: x["line"])))


class Pylint(Linter):
    requires_install = ["pylint"]

    def run(self, files: list) -> None:
        import pylint.lint

        args = ["-f", "json"] + files
        file = io.StringIO()
        with redirect_stdout(file):
            pylint.lint.Run(args, exit=False)
        file.seek(0)

        text = file.read()
        self.status_code = 1 if text.strip() else 0

        data = json.loads(text) if text.strip() else []
        self.hook(data)


class Pyroma(Linter):
    requires_install = ["pyroma"]

    def run(self, _: list) -> None:
        file = io.StringIO()
        with redirect_stdout(file), redirect_stderr(io.StringIO()):
            # Import pyroma here because it uses logging and sys.stdout
            import pyroma  # noqa pylint: disable=all
            pyroma.run('directory', '.')
        file.seek(0)

        text = file.read()

        lines = text.split("\n")

        lines.pop(0)
        lines.pop(0)

        data = {'modules': {}}

        module = lines.pop(0)[6:].strip()
        data['modules'][module] = []
        lines.pop(0)
        if len(lines) >= 6:
            line = lines.pop(0)
            while line != "-" * 30:
                data['modules'][module].append(line)
                line = lines.pop(0)

        rating = lines.pop(0)
        data['rating'] = int(rating[14:-3])
        data['rating_word'] = lines.pop(0)

        self.status_code = 0 if data['rating'] == 10 else 1

        if data['rating'] == 10:
            data = []

        self.hook(data)



"""
Linter/Styler/Checker combined for linters.

Linters supported:

- flake8
- flake8-bugbear (ext)
- flake8-import-order (ext)
- flake8-docstrings (ext)
- flake8-todo (ext)
- flake8-requirements (ext)
- flake8-string-format (ext)
- flake8-tidy-import (ext)
- flake8-bandit (ext, bandit)
- pylint
- yapf
- isort
- pyroma
- safety
- bandit
- dodgy
- vulture
- pytest
- Upload to pypi
"""

# Stdlib
import argparse

# External Libraries
from snekchek.lint import get_linters
from snekchek.secure import get_security
from snekchek.structure import CheckHandler
from snekchek.style import get_stylers
from snekchek.tool import get_tools


def run_main(args: argparse.Namespace, do_exit=True) -> None:
    """Runs the checks and exits.

    To extend this tool, use this function and set do_exit to False
    to get returned the status code.
    """
    handler = CheckHandler(
        file=args.config_file, out_json=args.json, files=args.files)

    for style in get_stylers():
        handler.run_linter(style())

    for linter in get_linters():
        handler.run_linter(linter())

    for security in get_security():
        handler.run_linter(security())

    for tool in get_tools():
        tool = tool()

        # Only run pypi if everything else passed
        if tool.name == "pypi" and handler.status_code != 0:
            continue

        handler.run_linter(tool)

    if do_exit:
        handler.exit()
    return handler.status_code


def main() -> None:
    """Main entry point for console commands."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--json",
        help="output in JSON format",
        action="store_true",
        default=False)
    parser.add_argument(
        "--config-file", help="Select config file to use", default=".snekrc")
    parser.add_argument(
        'files',
        metavar='file',
        nargs='*',
        default=[],
        help='Files to run checks against')

    args = parser.parse_args()

    run_main(args)


if __name__ == "__main__":
    main()



# External Libraries
from setuptools import setup, find_packages
from snekchek import misc
from snekchek.lint import get_linters
from snekchek.secure import get_security
from snekchek.style import get_stylers
from snekchek.tool import get_tools

with open("README.rst") as file:
    README = file.read()

with open("requirements.txt") as file:
    REQUIREMENTS = file.readlines()

EXTRAS = {}

for group in [get_linters(), get_security(), get_stylers(), get_tools()]:
    for linter in group:
        EXTRAS[linter.__name__.lower()] = linter.requires_install

if __name__ == '__main__':
    setup(
        name="snekchek",
        author="izunadevs",
        author_email="izunadevs@martmists.com",
        maintainer="martmists",
        maintainer_email="mail@martmists.com",
        license="MIT",
        zip_safe=False,
        version=misc.__version__,
        description=misc.description,
        long_description=README,
        url="https://github.com/IzunaDevs/SnekChek",
        packages=find_packages(),
        install_requires=REQUIREMENTS,
        extras_require=EXTRAS,
        entry_points={
            "console_scripts": ["snekchek = snekchek.__main__:main"]
        },
        keywords=[
            "lint", "python", "pylint", "flake8", "isort", "vulture", "bandit",
            "dodgy", "safety", "pytest", "pyroma", "pypi", "yapf", "snekchek",
            "snekrc"
        ],
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console", "Intended Audience :: Developers",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.3",
            "Programming Language :: Python :: 3.4",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Topic :: Software Development :: Quality Assurance",
            "Topic :: Software Development :: Libraries :: Python Modules"
        ],
        python_requires=">=3.0")



"""Base config for entry points."""

# Stdlib
import os

# Snekchek
import configobj

__all__ = ("config", )

config = configobj.ConfigObj({  # pylint: disable=invalid-name
    "all": {
        "linters": [
            "flake8", "pylint", "yapf", "isort", "pyroma", "safety", "dodgy",
            "vulture", "pytest", "pypi"
        ]
    },
    "pypi": {
        "TWINE_USERNAME": os.environ.get("TWINE_USERNAME", "abc"),
        "TWINE_PASSWORD": os.environ.get("TWINE_PASSWORD", "abc"),
        "sign": False,
        "TWINE_REPOSITORY": "pypi",
        "quiet": False
    },
    "flake8": {
        "max-line-length": 79,
        "exclude": ["build", "dist"],
        "ignore": [],
        "quiet": False
    },
    "pylint": {
        "quiet": False
    },
    "yapf": {
        "inplace": True,
        "quiet": False
    },
    "isort": {
        "line_length": 79,
        "inplace": True,
        "indent": '    ',
        "sections":
        "FUTURE,STDLIB,THIRDPARTY,FIRSTPARTY,LOCALFOLDER".split(","),
        "quiet": False
    },
    "bandit": {
        "quiet": False
    },
    "style": {
        "inplace": True,
        "quiet": False
    },
    "pyroma": {
        "quiet": False
    },
    "vulture": {
        "min-confidence": 60,
        "verbose": False,
        "exclude": [],
        "sort-by-size": False,
        "quiet": False
    },
    "safety": {
        "quiet": False,
        "ignore": [],
        "pyup_key": '',
        "db_path": ''
    },
    "dodgy": {
        "quiet": False,
        "ignore_paths": []
    }
})
